/*
 * Copyright (c) 1991-2003 Rational Software Corporation. All rights
 * reserved.
 * 
 * Rational, the Rational logo, ObjecTime (and logo) are registered
 * trademarks of Rational Software Corporation.
 * 
 * The placing of the copyright notices on Rational source code is not
 * intended to mean that they are published works, but are included in the
 * event of inadvertent publication. The source code of Rational contains
 * trade secrets and other valuable confidential information of Rational
 * Software Corporation. Rational Software Corporation and its licensors
 * retain ownership to Rational Products in both source and object code
 * format, and related documentation and all rights therein including
 * copyrights, patents and trade secrets.
 * 
 * 
 * U.S. Government Rights
 * 
 * Use, duplication, or disclosure by the U.S. Government is subject to
 * restrictions set forth in the applicable Rational License Agreement
 * and in DFARS 227.7202-1(a) and 227.7202-3(a) (1995), DFARS
 * 252.227-7013(c)(1)(ii)(Oct 1988), FAR 12.212(a) 1995, FAR 52.227-19,
 * or FAR 52.227-14, as applicable.
 * 
 * Unless otherwise agreed to in writing, use of Rational products is
 * governed by the Rational Product License Agreement provided with
 * the product.
 */

#ifdef PRAGMA
#pragma implementation "RTActor_class.h"
#pragma implementation "RTComponentDescriptor.h"
#endif

#include <RTActorRef.h>
#include <RTActorSlot.h>
#include <RTActor_class.h>
#include <RTComponentDescriptor.h>
#include <RTDiag.h>
#include <RTImportList.h>

// The _contents field is interpreted differently depending on the
// replication factor (n = _info->replication) and the kind of
// reference.
//
// The _contents field is always nil for excluded references. They are
// identified by their replication factor which is less than one.
//
// If a fixed or optional reference has replication factor 1, we just
// use _contents to point at the only possible member.
//
// For fixed references with larger replication factors _contents
// addresses an array of n pointers to actors. There is no free list
// required because the application may not incarnate or destroy
// individual actors in such a reference.
//
// Optional references are broken down into two more cases: small
// and large replication factors. This distinction is based on the
// assumption that pointers to actors will all be aligned at even
// addresss. We can then use the least significant bit to distinguish
// between a pointer and a word which represents an entry in our free
// list. The remaining bits of a pointer are divided into two integers
// which represent the indices of the previous and next entries in the
// free list. On machines where pointers occupy 4 bytes, the range of
// replication factors that can use this scheme is 2 through 32767,
// inclusive. This dual view is captured by the RTSmallOptional type.
//
// For optional references with replication factors that are too large
// for this strategy, we use the RTLargeOptional struct with two
// pointers for each entry.
//
// Whether we use RTSmallOptional or RTLargeOptional, we allocate
// one extra object which is used as the head of the free list.
//
// For imported references with replication factor 1, _contents points
// to a single RTImportList object. There is no need for a free list.
//
// Imported references with larger replication factor, we allocate
// a pair of pointers that serve as the head of the free list and array
// of n RTImportList objects. While an entry is in the free list the
// links in RTImportList::Unused are valid. When an entry is reserved,
// the reference field of RTImportList::Used is made to point at the
// containing reference. The Used and Unused structures share the same
// space: there is no possibility for confusion because the RTActorRef
// object lies outside the range of addresses to which the Unused links
// can refer. All elements of all these arrays are pointers, so there
// should be no alignment problems.

void RTActorRef::init( const RTComponentDescriptor * info )
{
    // this can only be called once
    RTAssert( _info == (const RTComponentDescriptor *)0 );

    _info = info;

    int n = info->replication;

    switch( info->kind )
    {
    case RTComponentDescriptor::Imported:
	if( n <= 0 )
	{
	    // excluded
	}
	else if( n == 1 )
	{
	    // We use new[] here so the destructor can always use delete[].
	    RTImportList * entry = new RTImportList [ 1U ];

	    entry->used.reference = (RTActorRef *)0;
	    entry->used.actor     = (RTActor    *)0;

	    _contents = entry;
	}
	else
	{
	    unsigned               pointers;
	    RTImportList::Unused * head;
	    RTImportList         * entry;

	    pointers  = sizeof( RTImportList::Unused ) + sizeof( void * ) - 1U;
	    pointers += (unsigned)n * sizeof( RTImportList );
	    pointers /= sizeof( void * );

	    _contents = new void * [ pointers ];

	    head  = (RTImportList::Unused *)_contents;
	    entry = (RTImportList *)(void *)&head[ 1 ];

	    head ->       next = entry;
	    entry->unused.prev = (RTImportList *)(void *)head;

	    --n;
	    do
	    {
		entry[ 0 ].  used.actor = (RTActor *)0;
		entry[ 0 ].unused.next  = &entry[ 1 ];
		entry[ 1 ].unused.prev  = &entry[ 0 ];
		++entry;
	    }
	    while( --n != 0 );

	    entry->  used.actor = (RTActor *)0;
	    entry->unused.next  = (RTImportList *)(void *)head;
	    head ->       prev  = entry;
	}
	break;

    case RTComponentDescriptor::Optional:
	if( n <= 1 )
	{
	    // _contents will point directly at the only possible actor
	}
	else if( RTSmallOptional::supports( n ) != 0 )
	{
	    RTSmallOptional * entry = new RTSmallOptional [ (unsigned)n + 1U ];

	    _contents = entry;

	    entry->insert( n, 1 );

	    for( int i = 0; ++entry, ++i < n; )
		entry->insert( i - 1, i + 1 );

	    entry->insert( n - 1, 0 );
	}
	else
	{
	    RTLargeOptional * entry;

	    entry = new RTLargeOptional [ (unsigned)n + 1U ];

	    _contents = entry;

	    entry[ 0 ].prev = &entry[ n ];
	    entry[ n ].next = &entry[ 0 ];

	    do
	    {
		entry[ 0 ].next = &entry[ 1 ];
		entry[ 1 ].prev = &entry[ 0 ];
		++entry;
	    }
	    while( --n != 0 );
	}
	break;

    case RTComponentDescriptor::Fixed:
	if( n <= 1 )
	{
	    // _contents will point directly at the only possible actor
	}
	else
	{
	    RTActor ** entry = new RTActor * [ (unsigned)n ];

	    _contents = entry;

	    do
		*entry++ = (RTActor *)0;
	    while( --n != 0 );
	}
	break;

    default:
	RTDiag::panic( "unexpected component kind" );
	//lint -unreachable
	break;
    }
}
