/*
 * Copyright (c) 1991-2003 Rational Software Corporation. All rights
 * reserved.
 * 
 * Rational, the Rational logo, ObjecTime (and logo) are registered
 * trademarks of Rational Software Corporation.
 * 
 * The placing of the copyright notices on Rational source code is not
 * intended to mean that they are published works, but are included in the
 * event of inadvertent publication. The source code of Rational contains
 * trade secrets and other valuable confidential information of Rational
 * Software Corporation. Rational Software Corporation and its licensors
 * retain ownership to Rational Products in both source and object code
 * format, and related documentation and all rights therein including
 * copyrights, patents and trade secrets.
 * 
 * 
 * U.S. Government Rights
 * 
 * Use, duplication, or disclosure by the U.S. Government is subject to
 * restrictions set forth in the applicable Rational License Agreement
 * and in DFARS 227.7202-1(a) and 227.7202-3(a) (1995), DFARS
 * 252.227-7013(c)(1)(ii)(Oct 1988), FAR 12.212(a) 1995, FAR 52.227-19,
 * or FAR 52.227-14, as applicable.
 * 
 * Unless otherwise agreed to in writing, use of Rational products is
 * governed by the Rational Product License Agreement provided with
 * the product.
 */

#include <RTDiag.h>
#include <RTFloat.h>
#include <RTFormat.h>
#include <RTObject_class.h>

#include <float.h>

// This sequence is carefully constructed to detect the smallest value
// such that the value+1 cannot be precisely represented in the mantissa.
// Then log2(value) is the number of bits of precision in the mantissa.
// There are two possible outcomes when we try to compute value+1: it gets
// rounded up or truncated. Either way the final value of test is non-zero.

inline static bool fits_mantissa( double value )
{
    double test = value + 1.0;

    // Make sure the compiler doesn't optimize away the addition
    // and subtraction of value. We know this call does nothing,
    // but the compiler must assume that it modifies test.

    RTType_void.init( &test );

    test -= value;

    return ( test == 1.0 ) ? true : false;
}

int RTFormat::_double( double value, char * buffer, int bufferSize )
{
    int size;

    if( value == 0.0 )
    {
	if( bufferSize < 2 )
	    size = 0;
	else
	{
	    buffer[ 0 ] = '0';
	    buffer[ 1 ] = '\0';
	    size = 1;
	}
    }
    else if( value < 0.0 )
    {
	if( (size = _double( - value, &buffer[ 1 ], bufferSize - 1 )) != 0 )
	{
	    buffer[ 0 ] = '-';
	    ++size; // account for sign
	}
    }
    else if( value > DBL_MAX )
    {
	if( bufferSize < 4 )
	    size = 0;
	else
	{
	    buffer[ 0 ] = 'I';
	    buffer[ 1 ] = 'n';
	    buffer[ 2 ] = 'f';
	    buffer[ 3 ] = '\0';
	    size = 3;
	}
    }
    else if( value > 0.0 ) // false only for +NaN which is handled below
    {
	static int MaxDigits = 0;

	if( MaxDigits == 0 )
	{
	    // We need log2( 10 ) > 3.32 bits to for one decimal digit of
	    // precision. Compute the number of decimal digits that can be
	    // represented in a double by testing the size of the mantissa.
	    // We do this to avoid loss of accuracy as we repeatedly
	    // multiply by two. The s100_ prefix denotes values which are
	    // scaled by one hundred.

	    enum
	    {
		   s100_log10       = 332
		 , s100_double_bits = 100 * 8 * (int)sizeof( double )
	    };
	    int    s100_bits        = s100_log10 - 1;
	    double power           =  1.0;

	    do
	    {
		// The mantissa is only a part of the
		// floating-point representation.
		RTAssert( s100_bits < s100_double_bits + s100_log10 - 1 );

		s100_bits += 100;
		power     *= 2.0;
	    }
	    while( fits_mantissa( power ) );

	    MaxDigits = s100_bits / s100_log10;
	}

	RTFloat conversion;

	conversion.assign( value );

	// room for leading '0.0' and trailing 'e+9999\0'
	char   decimal[ 3 + RTFloat::mantissa_digits + 8 ];

	char * end      = &decimal[ 3 + RTFloat::mantissa_digits ];
	int    exponent = conversion.exp10;
	char * cursor;
	int    digits;

	for( cursor = end, digits = 0; ; )
	{
	    *--cursor = (char)( '0' + conversion.mod10() );
	    ++digits;

	    if( conversion.div10() == 0 )
		break;
	}

	int excess = digits - MaxDigits;

	if( excess > 0 )
	{
	    // We have more digits than necessary to specify the value
	    // precisely. Unless we would be simply replacing those
	    // digits with zeros below, round and adjust the exponent.

	    if( exponent > 0 && excess + exponent <= 3 )
	    {
		// not too many: keep them in favour of zeros
	    }
	    else
	    {
		exponent += excess;
		digits   -= excess;
		end      -= excess;

		if( *end >= '5' )
		{
		    cursor[ -1 ] = '0';
		    while( *--end == '9' )
		    {
			++exponent;
			--digits;
		    }
		    *end++ += 1;

		    if( digits == 0 )
		    {
			digits = 1;
			--cursor;
		    }
		}

		// trim any trailing zeros
		while( *--end == '0' )
		{
		    ++exponent;
		    --digits;
		}
		++end;
	    }
	}

	switch( exponent )
	{
	default:
	{
	    if( exponent > 0 )
	    {
		if( exponent <= 3 )
		{
		    do
			*end++ = '0';
		    while( --exponent > 0 );
		    break;
		}
	    }
	    else
	    {
		int point = exponent + digits;

		if( point >= -1 )
		{
		    while( point <= 0 )
		    {
			*--cursor = '0';
			++point;
		    }

		    char * shift = --cursor;

		    do
			shift[ 0 ] = shift[ 1 ];
		    while( ++shift, --point > 0 );
		    *shift = '.';
		    break;
		}
	    }

	    if( digits != 1 )
	    {
		--cursor;
		cursor[ 0 ] = cursor[ 1 ];
		cursor[ 1 ] = '.';
	    }

	    exponent += digits - 1;

	    *end++ = 'e';
	    if( exponent > 0 )
		*end++ = '+';

	    int exp_size = _int( exponent, end, 6 );

	    if( exp_size == 0 )
		return 0;

	    end += exp_size;
	    break;
	}

	case 0:
	    break;
	}

	for( size = 0; cursor != end; ++size )
	{
	    if( --bufferSize <= 0 )
		return 0;

	    *buffer++ = *cursor++;
	}

	*buffer = '\0';
    }
    else
    {
	if( bufferSize < 4 )
	    size = 0;
	else
	{
	    buffer[ 0 ] = 'N';
	    buffer[ 1 ] = 'a';
	    buffer[ 2 ] = 'N';
	    buffer[ 3 ] = '\0';
	    size = 3;
	}
    }

    return size;
}
