/*
 * Copyright (c) 1991-2003 Rational Software Corporation. All rights
 * reserved.
 * 
 * Rational, the Rational logo, ObjecTime (and logo) are registered
 * trademarks of Rational Software Corporation.
 * 
 * The placing of the copyright notices on Rational source code is not
 * intended to mean that they are published works, but are included in the
 * event of inadvertent publication. The source code of Rational contains
 * trade secrets and other valuable confidential information of Rational
 * Software Corporation. Rational Software Corporation and its licensors
 * retain ownership to Rational Products in both source and object code
 * format, and related documentation and all rights therein including
 * copyrights, patents and trade secrets.
 * 
 * 
 * U.S. Government Rights
 * 
 * Use, duplication, or disclosure by the U.S. Government is subject to
 * restrictions set forth in the applicable Rational License Agreement
 * and in DFARS 227.7202-1(a) and 227.7202-3(a) (1995), DFARS
 * 252.227-7013(c)(1)(ii)(Oct 1988), FAR 12.212(a) 1995, FAR 52.227-19,
 * or FAR 52.227-14, as applicable.
 * 
 * Unless otherwise agreed to in writing, use of Rational products is
 * governed by the Rational Product License Agreement provided with
 * the product.
 */

#include <RTActor.h>
#include <RTActorProbe.h>
#include <RTController.h>
#include <RTDebugger.h>
#include <RTDiag.h>
#include <RTMessage.h>
#include <RTPortDescriptor.h>
#include <RTProtocol.h>
#include <RTProtocolDescriptor.h>
#include <RTSpecialPorts.h>
#include <RTSpecialSignals.h>
#include <RTTimerNode.h>

RTMessage::DeliverResult RTMessage::deliver( void )
{
    RTActor      * destination = _receiver;
    RTController * destContext = destination->context();
    int            id;

    if( _port == (RTProtocol *)0 )
    {
	id = (int)RTControlPort;
	switch( (int)signal )
	{
	  case RTInitSignal:
	    // some other thread asked to init an actor
	    destination->msg = this;

#if LOG_MESSAGE
	    destination->logMsg();
#endif

	    destination->initializeAll();
	    destination->processHistory();

#if OBSERVABLE
	    if( destination->probe != (RTActorProbe *)0 )
		destination->probe->evtChain( 0 );
#endif

	    // message was delivered, now clear msg in destination
	    destination->msg = (RTMessage *)0;

	    return Accepted;

	  case RTDestroySignal:
	    // some other thread asked to destroy an actor
	    {
		RTActor * owner = *(RTActor **)data; //lint !e413 !e1402

		RTAssert( owner != (RTActor *)0 );

		destContext->destroy( destination, owner );
		if( destContext->send( owner, (int)RTAckSignal,
					      data, (const RTObject_class *)0,
					      (int)Synchronous ) == 0 )
		    RTDiag::panic( "Can't reply to destroy signal" );
	    }
	    return Accepted;

	  case RTAckSignal:
	    // This is an acknowledgement: the request sent by 'receiver'
	    // or on its behalf has been completed. Clearing the pointer
	    // addressed by data will accept this acknowledgement
	    *(void **)data = (void *)0; //lint !e413
	    return Accepted;

	  case RTTimeoutSignal:
	    return RTTimerNode::deliver( this );

	  case RTPendingTimeoutSignal:
	    // This message should not have been queued!
	    RTDiag::panic( "Queued pending timeout" );
	    //lint -unreachable
	    return Accepted;

#if OBSERVABLE
	  case RTProbeSignal:
	    // probe wants attention, possibly ESV read/write
	    destContext->debugger()->probeAttention( destination );
	    return Accepted;
#endif

	  case RTInterruptSignal:
	    break;

	  case RTStartSignal:
	  case RTStopSignal:
	    RTDiag::panic( "Remote start and stop not yet implemented" );
	    //lint -unreachable
	    return Accepted;

	  case RTWakeupSignal:
	    // This was just for waking up the message-receiving-thread.
	    // The thread has already been awaken. Now is the time to
	    // gracefully handle this signal.
	    return Accepted;

	  default:
	    RTDiag::panic( "There is no such control message!" );
	    //lint -unreachable
	    return Accepted;
	}
    }
    else
	id = _port->getId();

#if SUSPEND
    if( destination->isSuspended() )
    {
	next = destination->suspendMsgs;
	destination->suspendMsgs = this;
	return Deferred;
    }
#endif

    destination->msg = this;

#if OBSERVABLE
    RTActorProbe * probe = destination->probe;

    if( probe != (RTActorProbe *)0 )
    {
	probe->evtMessageIn( this );

	if( destContext->getReplyBuffer() != (RTMessage *)0 )
	    probe->evtMessageOutIn( this, RTEventInfo::SendInvoke );
	else
	    probe->evtMessageOutIn( this, RTEventInfo::SendNormal );
    }
#endif

#if LOG_MESSAGE
    destination->logMsg();
#endif

    RTMessage::DeliverResult result = Accepted;

#if RTTYPECHECK_RECEIVE != RTTYPECHECK_DONT
    if( _port == (RTProtocol *)0 )
    {
    }
    else switch( _port->receiveCheck( (int)signal, _type ) )
    {
    case RTProtocolDescriptor::badType:
	destContext->setError( RTController::badClass,
			       RTController::CommDeliver,
			       this,
			       (const void *)0 );
#if RTTYPECHECK_RECEIVE == RTTYPECHECK_FAIL
	result = Rejected;
#endif
	break;

    case RTProtocolDescriptor::badSignal:
	destContext->setError( RTController::badSignal,
			       RTController::CommDeliver,
			       this,
			       (const void *)0 );
#if RTTYPECHECK_RECEIVE == RTTYPECHECK_FAIL
	result = Rejected;
#endif
	break;

    case RTProtocolDescriptor::ok:
    default:
	break;
    }
#endif // RTTYPECHECK_RECEIVE != RTTYPECHECK_DONT

#if RTTYPECHECK_RECEIVE == RTTYPECHECK_FAIL
    if( result == Accepted )
#endif
    {
	destination->rtsBehavior( (int)signal, id );
    }

#if OBSERVABLE
    if( probe != (RTActorProbe *)0 )
	probe->evtChain( 0 );
#endif

    // message was delivered, now clear msg in destination
    destination->msg = (RTMessage *)0;

    if( isDeferred() != 0 )
    {
#if DEFER_IN_ACTOR
	destination->deferredQ.nqTail( this );
#else
	destContext->deferredQ.nqTail( this );
#endif
	result = Deferred;
    }

    return result;
}
