#! perl -w

# Copyright (c) 1991-2003 Rational Software Corporation. All rights
# reserved.
# 
# Rational, the Rational logo, ObjecTime (and logo) are registered
# trademarks of Rational Software Corporation.
# 
# The placing of the copyright notices on Rational source code is not
# intended to mean that they are published works, but are included in the
# event of inadvertent publication. The source code of Rational contains
# trade secrets and other valuable confidential information of Rational
# Software Corporation. Rational Software Corporation and its licensors
# retain ownership to Rational Products in both source and object code
# format, and related documentation and all rights therein including
# copyrights, patents and trade secrets.
# 
# 
# U.S. Government Rights
# 
# Use, duplication, or disclosure by the U.S. Government is subject to
# restrictions set forth in the applicable Rational License Agreement
# and in DFARS 227.7202-1(a) and 227.7202-3(a) (1995), DFARS
# 252.227-7013(c)(1)(ii)(Oct 1988), FAR 12.212(a) 1995, FAR 52.227-19,
# or FAR 52.227-14, as applicable.
# 
# Unless otherwise agreed to in writing, use of Rational products is
# governed by the Rational Product License Agreement provided with
# the product.

# This script performs two functions: It translates object file lists
# (identified by %file on the command line) to linker command files,
# and adds the constructor/destructor tables to the executable.
#
# Example usage:
# ===============
# perl link.pl ARCH=386 -o TopCapsule \
#   main.o %OLIST user_obj.o -LX -lx ...
#
# The general algorithm:
#   Translate each OLIST to OLIST.cmd
#   ld386 -r -o RTmunch.o main.o -T OLIST.cmd user_obj.o -LX -lx
#   nm386 RTmunch.o | munch > RTctor.c
#   cc386 -traditional -o RTctor.o -c RTctor.c
#   ld386 -r -o TopCapsule RTctor.o RTmunch.o

# Win9x command.com does not interact with perl according to win32 standard.
# We therefore need to call some functions in a non-standard way.
#
# This detects if the system is Win95 for that purpose.

sub mySystem
{
    my( $OS, $ROSERT_HOST );

    $OS          = $ENV{'OS'}          || '?';
    $ROSERT_HOST = $ENV{'ROSERT_HOST'} || '?';

    if( $ROSERT_HOST eq 'win32' && $OS !~ m{^Windows_NT$}i )
    {
	return system( join( ' ', @_ ) );
    }
    else
    {
	return system( @_ );
    }
}

sub report_ges_error
{
    my( $ERROR ) = @_;

    print "description '$ERROR' severity 'error'\n";
    print "result 'failure'\n";

    exit 1;
}

sub show_usage
{
    &report_ges_error( "Usage: $0 ARCH=<arch> options ... " );
}

&show_usage() unless $#ARGV >= 1;

if( $ARGV[0] =~ m{^ARCH=(\w+)$} )
{
    $arch = $1;
    shift @ARGV;
}
else
{
    &show_usage();
}

&report_ges_error( "WIND_BASE is not set" )
	unless( $WIND_BASE = $ENV{'WIND_BASE'} );

&report_ges_error( "WIND_HOST_TYPE is not set" )
	unless( $WIND_HOST_TYPE = $ENV{'WIND_HOST_TYPE'} );

# commands
$cc_cmd  = "cc$arch";
$ld_cmd  = "ld$arch";
$munch   = 'munch';
$nm_cmd  = "nm$arch --extern-only --no-sort";

# filenames
$prelink = 'RTmunch.o';
$ctor_c  = 'RTctor.c';
$ctor_o  = 'RTctor.o';
$output  = 'a.out';

$status   = 0;                              # assume all is well
$cmdfile  = 'ld000';                        # command filename pattern
@tmpfiles = ( $prelink, $ctor_c, $ctor_o ); # temporary files

# Translate all indirect object file lists (identified by %) into linker
# command files for use in the prelink step.

@command  = ( $ld_cmd, '-r', '-o', $prelink );

if( $arch eq 'ppc' )
{
    @script_intro = ( '-T', "$WIND_BASE/host/$WIND_HOST_TYPE"
			  . '/powerpc-wrs-vxworks/lib/ldscripts/elf32ppc.xr' );
}
else
{
    @script_intro = ();
}

foreach $arg ( @ARGV )
{
    if( $arg eq '-o' )
    {
	$lasto = 1;
    }
    elsif( $lasto )
    {
	$output = $arg;
	$lasto  = 0;
    }
    elsif( $arg =~ m{^%(.*)$} )
    {
	$in = $1;
	if( ! open( IN, "<$in" ) )
	{
	    print STDERR "link.pl: Can't read '$in': $!\n";
	    $status = 1;
	}
	else
	{
	    $out = "$cmdfile.cmd";
	    ++$cmdfile;
	    if( ! open( OUT, ">$out" ) )
	    {
		print STDERR "link.pl: Can't create '$out': $!\n";
		$status = 1;
	    }
	    else
	    {
		while( <IN> )
		{
		    chomp;
		    print OUT "INPUT(", $_, ".o)\n";
		}
		close( OUT );

		push @command, @script_intro, '-T', $out;
		@script_intro = ();
		push @tmpfiles, $out;
	    }
	    close( IN );
	}
    }
    else
    {
	push @command, $arg;
    }
}

# prelink everything

$status = &mySystem( @command ) if $status == 0;

# munch the big object file

$status = &mySystem( "$nm_cmd $prelink | $munch > $ctor_c" )
				if $status == 0;

# compile the constructor/destructor tables

$status = &mySystem( $cc_cmd, '-traditional', '-o', $ctor_o, '-c', $ctor_c )
				if $status == 0;

# final link of the application

$status = &mySystem( $ld_cmd, '-r', '-o', $output, $ctor_o, $prelink )
				if $status == 0;

# remove our temporary files

unlink( @tmpfiles );

exit $status if $status < 256;
exit $status / 256;
